import os
import glob
from yaml import load
from os import link
from pathlib import Path


def autolink(config):
    S = Path('Scripts')
    tasks = load(open(config))

    for filename in glob.iglob('Scripts/**/*.ln.R', recursive=True):
        os.remove(filename)

    for task in tasks:
        print(task)
        if task['dst'] is None:
            continue

        if task['src'] is None:
            continue

        for dst in task['dst']:
            if dst is None:
                continue
            if not os.path.exists(str(S / dst)):
                os.makedirs(str(S / dst))
            for src in task['src']:
                if src is None:
                    continue

                link(str(S / Path(src)), str(S / Path(dst) / Path(src).stem) + '.ln.R')



import os
import sys
import pathlib
import re
from wbuild.utils import getWBData, getMDData, getYamlParam
sys.path.insert(0, os.getcwd() + "/.wBuild")

htmlPath = "Output/html"

WB_FIELDS = {"type"}
# SNAKEMAKE  = ["input", "output", "threads"]
SNAKEMAKE_FIELDS = ["input",
                    "output",
                    "params",
                    "threads",
                    "resources",
                    "priority",
                    "version",
                    "log",
                    "message",
                    "run",
                    "shell",
                    "script"]


def writeDependencyFile():
    print("Updating dependencies...")
    wbData = getWBData()
    mdData = getMDData()
    with open('.wBuild.depend', 'w') as f:
        f.write('######\n')
        f.write('#This is a autogenerated snakemake file by wBuild\n')
        f.write('#wBuild by Leonhard Wachutka\n')
        f.write('######\n')

        # write rules
        for r in wbData:
            writeRule(r, f)
        # write md rules
        for r in mdData:
            writeMdRule(r, f)

        # write build index rule
        writeIndexRule(wbData, mdData, f)
    print("Done.\n")


def joinEmpty(string_list):
    # filter empty string list
    return ", ".join([x for x in string_list if x.strip() != ''])


def escapeSMString(item):
    if type(item) is dict:
        return str(list(item.keys())[0]) + ' = ' + escapeSMString(str(list(item.values())[0]))

    elif type(item) is str:
        if item.startswith("`sm ") and item.endswith("`"):
            return item[4:-1]
        return "'" + item + "'"
    return ''


def ensureString(elem):
    if elem is None:
        return ''
    elif type(elem) is list:
        if len(elem) == 0:
            return ''
        else:
            # make sure each element is a character
            elem = [escapeSMString(item) for item in elem]
            elem = [x for x in elem if str(x).strip() != '']

            return ", ".join(elem)
    elif type(elem) is str:
        if "," not in elem:
            return "'" + elem + "'"
        else:
            return elem
    else:
        raise TypeError("Don't know how to handle type: " + str(type(elem)))


def dumpSMRule(dumpDic, file, sFile):
    if 'py' in dumpDic:
        code = dumpDic['py']
        if type(code) is str:
            file.write(insertPlaceholders(code, sFile))
        elif type(code) is list:
            [file.write(insertPlaceholders(line, sFile) + '\n') for line in code]

    file.write('rule ' + dumpDic['rule'] + ':\n')
    for field in SNAKEMAKE_FIELDS:
        if field in dumpDic:
            file.write('    ' + field + ': ' + str(dumpDic[field]) + '\n')


def insertPlaceholders(s, file):
    path = pathlib.Path(file)
    PD = pathlib.Path('Output/ProcessedData')

    PP = path.parts[-2]
    s = s.replace("{wbPD}", str(PD))
    s = s.replace("{wbPP}", str(PP))

    if len(path.parts) <= 2 and bool(re.search('{wbP(D_P*)?}', file)):
        print("If using placeholders please make sure you have the right",
              " directory structure.")

    if len(path.parts) > 2:
        P = path.parts[-3]
        s = s.replace("{wbPD_P}", str(PD / P))
        s = s.replace("{wbPD_PP}", str(PD / P / PP))
        s = s.replace("{wbP}", str(P))

    return s


def writeRule(r, file):
    # file needs wb to write to Snakefile
    if "wb" not in r["param"] or type(r["param"]['wb']) is not dict:
        return
    elem = r["param"]["wb"]

    # extract rule
    # rule = r['file'].replace('.','_').replace('/','_')

    # determine input, output and script
    elem["input"] = insertPlaceholders(joinEmpty([ensureString(elem.get("input")), "RScript = '" + r['file'] + "'"]), r['file'])
    if elem.get("type") == 'script':
        elem["output"] = insertPlaceholders(ensureString(elem.get("output")), r['file'])
        elem["script"] = '\'' + r['file'] + '\''
    elif elem.get("type") == 'noindex':
        elem["output"] = insertPlaceholders(ensureString(elem.get("output")), r['file'])
        elem["script"] = "'.wBuild/wBRender.R'"
    else:
        elem["output"] = insertPlaceholders(joinEmpty([ensureString(elem.get("output")), "wBhtml = '" + r['outputFile'] + "'"]), r['file'])
        elem["script"] = "'.wBuild/wBRender.R'"

    # remove wb related elements
    # elem = {key: elem[key] for key in elem if key not in WB_FIELDS}
    # if not set(elem.keys()).issubset(SNAKEMAKE_FIELDS):
    #    Warning("File: {0}. The following fields don't correspond to any snakemake or wBuild tag: {1}"
    #            .format(r['file'], ",".join(set(elem.keys()).difference(SNAKEMAKE_FIELDS))))

    # remove fields not in SNAKEMAKE_FIELDS
    # elem = {key: elem[key] for key in elem if key in SNAKEMAKE_FIELDS}
    elem['rule'] = r['file'].replace('.', '_').replace('/', '_')
    # write to file
    file.write('\n')
    # dumpDict = {'rule ' + rule: elem}
    # file.write(yaml.dump(dumpDict, default_flow_style = False, indent=4).replace("\'\'\'", "'").replace("\'\'", "'"))
    dumpSMRule(elem, file, r['file'])
    file.write('\n')


def writeMdRule(r, file):
    file.write('\n')
    file.write('rule ' + r['file'].replace('.', '_').replace('/', '_') + ':\n')
    file.write('    input: "' + r['file'] + '"\n')
    file.write('    output: "' + r['outputFile'] + '"\n')
    file.write('    shell: "pandoc --from markdown --to html --css .wBuild/lib/github.css --toc --self-contained -s -o {output} {input}"\n')

    file.write('\n')


def writeIndexRule(rows, mdrows, file):
    inputFiles = []
    for r in rows:
        if getYamlParam(r, 'type') == 'script' or getYamlParam(r, 'type') == 'noindex':
            continue
        inputFiles.append(r['outputFile'])

    for r in mdrows:
        inputFiles.append(r['outputFile'])

    file.write('\n')
    file.write('rule Index:\n')
    file.write('    input: \n        "' + '",\n        "'.join(inputFiles) + '"\n')
    file.write('    output: "' + htmlPath + '/index.html"\n')
    # file.write('    script: ".wBuild/createIndex.py"\n')
    file.write('    run:\n')
    file.write('        import wbuild.createIndex\n')
    file.write('        wbuild.createIndex.ci()\n')

    file.write('\n')


if __name__ == "__main__":
    writeDependencyFile()



# -*- coding: utf-8 -*-
"""CLI interface to wbuild."""

import sys
import os
import click
import wbuild
import pathlib
import shutil
import distutils.dir_util
import click_log
import logging
logger = logging.getLogger(__name__)
click_log.basic_config(logger)

def setup_paths():
    """Setup the wbuild paths
    """
    templatePath = pathlib.Path(wbuild.__file__).parent / 'template'
    wbuildPath = pathlib.Path(wbuild.__file__).parent / '.wBuild'
    demoPath = pathlib.Path(wbuild.__file__).parent / 'demo'
    return templatePath, wbuildPath, demoPath


@click.group()
@click_log.simple_verbosity_option(logger)
@click.version_option('1.1.4',prog_name='wBuild')
def main():
    pass


# --------------------------------------------
# commands

@main.command()
def init():
    """Initialize the repository with wbuild.

    This will create a .wBuild/ folder in the current path
    """
    if os.path.exists(".wBuild"):
        logger.error("ERROR: .wBuild already exists. Use update if you want to update the version")
        sys.exit(2)
    templatePath, wbuildPath, demoPath = setup_paths()
    distutils.dir_util.copy_tree(str(wbuildPath), './.wBuild')
    shutil.copy(str(templatePath / 'Snakefile'), '.')
    shutil.copy(str(templatePath / 'wbuild.yaml'), '.')
    shutil.copy(str(templatePath / 'readme.md'), '.')

    logger.info("init...done")


@main.command()
def demo():
    """Setup a demo wBuild demo project
    """
    if os.path.exists(".wBuild"):
        logger.error("ERROR: .wBuild already exists. Run demo in empty folder.")
        sys.exit(2)
    templatePath, wbuildPath, demoPath = setup_paths()
    shutil.copy(str(demoPath / 'Snakefile'), '.')
    shutil.copy(str(templatePath / 'wbuild.yaml'), '.')
    shutil.copy(str(templatePath / 'readme.md'), '.')
    distutils.dir_util.copy_tree(str(wbuildPath), './.wBuild')
    distutils.dir_util.copy_tree(str(demoPath), '.')
    logger.info("demo...done")


@main.command()
def update():
    """Update the .wBuild folder to the most recent version of wBuild
    """
    # - check if .wBuild exists
    #   - if it doesn't, return the error
    # - if it does, update it
    if not os.path.exists(".wBuild"):
        raise ValueError(".wBuild doesn't exists. Please run wBuild init first or move to the right directory")

    logger.info("Removing .wBuild")
    shutil.rmtree(".wBuild")
    logger.info("Running .Init")
    templatePath, wbuildPath, demoPath = setup_paths()
    distutils.dir_util.copy_tree(str(wbuildPath), './.wBuild')
    logger.info("update...done")



"""A list of synthax checkers
"""


def no_tabs(header):
    """Check that the header contains no tabs

    Args:
      header: Header string

    Returns:
      everything ok: empty string
      error: error message
    """
    if "\t" in header:
        return "Tabs found in the header. Please remove them"
    else:
        return ""


# list all the synthax checkers here
CHECKERS = [("no_tabs", no_tabs)]


def checkHeaderSynthax(header):
    """Check the header synthax

    Args:
      file_path: file to be checked
    """

    errors = ""
    for name, checker in CHECKERS:
        error = checker(header)
        if error:
            errors += "{0}: {1}".format(name, error)

    if errors:
        raise ValueError("Errors parsing the header file: \n{0}\n{1}".format(header, errors))



import os
import sys
import pathlib
import shutil
from string import Template
from os import listdir
from os.path import isfile, join
from wbuild.utils import getWBData, getMDData, getYamlParam

sys.path.insert(0, os.getcwd() + "/.wBuild")  # TODO - is this line required?


htmlPath = "Output/html"


def writeSubMenu(top, wbData, level):
    menuString = ''
    temp = []
    newWb = []
    for r in wbData:
        if (pathlib.PurePath(r['file']).parts[level - 1] == top):
            # Is it a file
            if(len(pathlib.PurePath(r['file']).parts) == (level + 1)):
                if getYamlParam(r, 'type') != 'script' and getYamlParam(r, 'type') != 'noindex':
                    menuString += '<li><a href="javascript:navigate(\'' + pathlib.PurePath(r['outputFile']).name + '\');">' + pathlib.PurePath(r['file']).parts[level] + '</a></li>\n'
                continue
            temp.append(pathlib.PurePath(r['file']).parts[level])
            newWb.append(r)
    temp = sorted(set(temp))
    for top in temp:
        menuString += '<li class="dropdown-submenu">\n'
        menuString += '    <a tabindex="-1" href="#">' + top + '</a>\n'
        menuString += '    <ul class="dropdown-menu">\n'
        menuString += '    ' + writeSubMenu(top, newWb, level + 1)
        menuString += '    </ul>\n'
        menuString += '</li>\n'
    return menuString


def getRecentMenu():
    menuString = ""
    htmlPath = "Output/html"
    rFiles = sorted([join(htmlPath, f) for f in listdir(htmlPath) if isfile(join(htmlPath, f))], key=os.path.getmtime, reverse=True)[:10]
    for f in rFiles:
        fo = pathlib.PurePath(f).name
        menuString += '<p><a href="javascript:navigate(\'' + fo + '\');">' + fo.replace('_', ' ').replace('.html', '') + '</a></p>\n'
    return menuString


def writeMenu():
    wbData = getWBData()
    mdData = getMDData()
    wbData += mdData
    menuString = ""
    temp = []
    for r in wbData:
        temp.append(pathlib.PurePath(r['file']).parts[1])
    temp = sorted(set(temp))
    for top in temp:
        menuString += '<li class="dropdown">\n'
        menuString += '   <a href="#" class="dropdown-toggle" data-toggle="dropdown" role="button" aria-haspopup="true" aria-expanded="false">' + top + '<span class="caret"></span></a>\n'
        menuString += '   <ul class="dropdown-menu multi-level" role="menu">\n'
        menuString += writeSubMenu(top, wbData, 2)
        menuString += '   </ul>\n'
        menuString += '</li>\n'
    template = open('.wBuild/template.html').read()
    template = Template(template).substitute(menu=menuString, title='title', rf=getRecentMenu())  # snakemake.config['projectTitle']

    f = open('Output/html/index.html', 'w')
    f.write(template)
    f.close()


def ci():
    writeMenu()

    libDir = "Output/html/lib"

    if os.path.exists(libDir):
        shutil.rmtree(libDir)
    shutil.copytree('.wBuild/lib', libDir)



import fnmatch
import os
import yaml
import operator
from functools import reduce
from wbuild.syntaxCheckers import checkHeaderSynthax


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def checkFileName(filename):
    """List of checks for the correct filename
    """
    if " " in filename:
        raise ValueError("Space not allowed in the filenames. File: {0}".filename)
    if "-" in os.path.basename(filename):
        raise ValueError("- not allowed in the filenames. File: {0}".filename)


def findFilesPath(path, patterns):
    """Recursively search for files following a certain pattern
    """
    matches = []
    for root, dirnames, filenames in os.walk(path):
        dirnames[:] = [d for d in dirnames if not d[0] == '_']
        dirnames[:] = [d for d in dirnames if not d[0] == '.']
        for filename in reduce(operator.add, (fnmatch.filter(filenames, p) for p in patterns)):
            checkFileName(filename)
            matches.append(os.path.join(root, filename))
    return sorted(matches)


def getSpinYaml(file):
    """Retrieves the yaml header (including)
    """
    yamlHeader = []
    for i, line in enumerate(open(file)):
        # first line has to start with #'---
        if i == 0 and not line.startswith("#'---"):
            raise ValueError("First line of the file should be equal to \"#'---\" ")

        # process
        li = line.strip()
        if li.startswith("#'"):
            yamlHeader.append(li[2:])

        # terminate on the second occurence of #'---
        if i != 0 and line.startswith("#'---"):
            break

    return '\n'.join(yamlHeader)


def checkYamlHeader(file):
    """Check the yaml header for synthax checks
    """
    with open(file, "r") as f:
        line = f.readline()
    if(line.startswith("#'---")):
        return True
    return False


def getWBData(script_dir="Scripts", htmlPath="Output/html"):
    """Parse all the R files

    Args:
      script_dir: Relative path to the Scripts directory
      htmlPath: Relative path to the html output path

    Returns:
      a list of dictionaries with fields:
      - file - what is the input R file
      - outputFile - there to put the output html file
      - param - parsed yaml header
    """
    out = []
    error = False
    for f in findFilesPath(script_dir, ['*.r', '*.R']):
        try:
            if not checkYamlHeader(f):
                # Ignore files not starting with
                continue
            header = getSpinYaml(f)
            # run all the synthax checks - will raise an error if it fails
            checkHeaderSynthax(header)
            param = next(yaml.load_all(header))
        except Exception as e:
            if not error:
                print(bcolors.FAIL + bcolors.BOLD + 'Could not parse', f,
                      '. Include valid yaml header. Not showing any further errors. \n',
                      'Errors {0}'.format(e) + bcolors.ENDC)
                error = True
            continue
        if('wb' in param):
            outFile = htmlPath + "/" + os.path.splitext(f)[0].replace('/', '_') + ".html"
            out.append({'file': f, 'outputFile': outFile, 'param': param})
    if error:
        raise ValueError("Errors occured in parsing the R files. Please fix them.")
    return out


def getMDData(script_dir="Scripts", htmlPath="Output/html"):
    """Parse all the .md files

    Args:
      script_dir: Relative path to the Scripts directory
      htmlPath: Relative path to the html output path

    Returns:
      a list of dictionaries with fields:
      - file - what is the input .md file
      - outputFile - there to put the output html file
      - param - parsed yaml header - always an empty list
    """
    out = []
    for f in findFilesPath(script_dir, ['*.md']):
        outFile = htmlPath + "/" + os.path.splitext(f)[0].replace('/', '_') + ".html"
        out.append({'file': f, 'outputFile': outFile, 'param': []})
    return out


def getYamlParam(r, paramName):
    if 'wb' in r['param'] and type(r['param']['wb']) is dict and paramName in r['param']['wb']:
        return r['param']['wb'][paramName]
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wbuild` package."""

from click.testing import CliRunner
from wbuild import cli


def test_help():
    """Test the cli help functions"""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help' in help_result.output
    assert 'Show this message and exit.' in help_result.output


def test_init(testdirectory):
    """Test the cli help functions"""
    init_dir = testdirectory.mkdir("init")
    r = init_dir.run("wbuild init")
    assert r.stdout.match('*init...done*')


def test_demo(testdirectory):
    """Test the cli help functions"""
    init_dir = testdirectory.mkdir("demo")
    r = init_dir.run("wbuild demo")
    assert r.stdout.match('*demo...done*')

    r = init_dir.run("snakemake")
